import os


def render_context(engine="pyglet", **kwargs):
    """
    Return the appropriate rendering context.

    At present, this accepts "pyglet", "osmesa" or "egl"

    Parameters
    ----------
    engine: str, "pyglet", "osmesa" or "egl"

    Returns
    -------
    RenderingContext

    """

    # PYOPENGL_PLATFORM must be set before any opengl imports
    if engine in ("osmesa", "egl"):
        os.environ["PYOPENGL_PLATFORM"] = engine

    import OpenGL.error

    if engine == "pyglet":
        from .pyglet_context import PygletRenderingContext

        try:
            return PygletRenderingContext(**kwargs)
        except OpenGL.error.Error as oee:
            msg = str(oee)
            if "no valid context" in msg:
                extramsg = (
                    "It looks like you have encountered an OpenGL context error while trying to start the GUI. "
                    "If you are running headless, try specifying 'osmesa' or 'egl' with the engine argument. If you "
                    "are running a newer Ubuntu (21+) or Fedora (34+) release, you may need to enforce Xorg usage "
                    "in PyOpenGL: see https://yt-idv.readthedocs.io/en/latest/installation.html#extra-steps-for-linux"
                )
                raise Exception(extramsg) from oee
            raise oee
    elif engine == "osmesa":
        from .osmesa_context import OSMesaRenderingContext

        return OSMesaRenderingContext(**kwargs)
    elif engine == "egl":
        from .egl_context import EGLRenderingContext

        return EGLRenderingContext(**kwargs)
    else:
        raise KeyError



import numpy as np
import traitlets
from yt.data_objects.data_containers import YTDataContainer
from yt.units.dimensions import length

from yt_idv.opengl_support import VertexArray, VertexAttribute
from yt_idv.scene_data.base_data import SceneData


class ParticlePositions(SceneData):
    name = "particle_positions"
    data_source = traitlets.Instance(YTDataContainer)
    particle_type = traitlets.Unicode("all")
    radius_field = traitlets.Unicode(None, allow_none=True)
    color_field = traitlets.Unicode(None, allow_none=True)
    position_field = traitlets.Unicode("particle_position")
    size = traitlets.CInt(-1)

    @traitlets.default("vertex_array")
    def _default_vertex_array(self):
        model_vertex = np.array(
            [[-1, 1], [-1, -1], [1, 1], [1, -1]], order="F", dtype="f4"
        )
        va = VertexArray(name="particle_positions")
        va.attributes.append(
            VertexAttribute(name="model_vertex", data=model_vertex, divisor=0)
        )
        for attr in ("position_field", "radius_field", "color_field"):
            if getattr(self, attr) is None:
                continue
            field = self.data_source[self.particle_type, getattr(self, attr)]
            if field.units.dimensions is length:
                field.convert_to_units("unitary")
            field = field.astype("f4").d
            if field.ndim == 1:
                field.shape = (field.size, 1)
            else:
                self.size = field.shape[0]  # for positions
            print(f"Setting {attr} to a field of shape {field.shape}")
            va.attributes.append(VertexAttribute(name=attr, data=field, divisor=1))
        print(f"Size is now: {self.size}")
        return va



import numpy
from setuptools import setup

from Cython.Build import cythonize  # isort:skip

_include_dirs = [numpy.get_include()]

setup(ext_modules=cythonize("yt_idv/utilities/*.pyx"), include_dirs=_include_dirs)



import traitlets
from OpenGL import GL

from yt_idv.scene_components.base_component import SceneComponent
from yt_idv.scene_data.line import LineData
from yt_idv.scene_data.rgba import RGBAData


class RGBADisplay(SceneComponent):
    name = "rgba_display"
    data = traitlets.Instance(RGBAData)
    priority = 20

    def _set_uniforms(self, scene, shader_program):
        shader_program._set_uniform("cm_tex", 0)

    def draw(self, scene, program):
        with self.data.colormap_texture.bind(0):
            GL.glDrawArrays(GL.GL_TRIANGLES, 0, 6)


class RGBALinePlot(SceneComponent):
    name = "rgba_line_plot"
    data = traitlets.Instance(LineData)
    priority = 20

    def draw(self, scene, program):
        for i, _channel in enumerate("rgba"):
            program._set_uniform("channel", i)
            GL.glDrawArrays(GL.GL_LINE_STRIP, 0, self.data.n_vertices)

    def _set_uniforms(self, scene, shader_program):
        pass



import pytest

from yt_idv.shader_objects import PreprocessorDefinitionState


def test_preprocessor_definition_state():

    pds = PreprocessorDefinitionState()

    pds.add_definition("fragment", ("USE_DB", ""))
    assert ("USE_DB", "") in pds["fragment"]
    pds.add_definition("vertex", ("placeholder", ""))
    assert ("placeholder", "") in pds["vertex"]

    with pytest.raises(ValueError, match="shader_type must be"):
        pds.add_definition("not_a_shader_type", ("any_str", ""))

    pds.clear_definition("fragment", ("USE_DB", ""))
    assert ("USE_DB", "") not in pds["fragment"]

    pds.reset("vertex")
    assert len(pds.vertex) == 0

    pds.add_definition("fragment", ("USE_DB", ""))
    pds.add_definition("geometry", ("placeholder", ""))
    pds.add_definition("vertex", ("placeholder", ""))
    pds.reset()
    for shadertype in pds._valid_shader_types:
        assert len(pds._get_dict(shadertype)) == 0



import numpy as np
import traitlets
from OpenGL import GL

from yt_idv.scene_annotations.base_annotation import SceneAnnotation
from yt_idv.scene_data.text_characters import TextCharacters

# This is drawn in part from
#  https://learnopengl.com/#!In-Practice/Text-Rendering


class TextAnnotation(SceneAnnotation):
    name = "text_annotation"
    data = traitlets.Instance(TextCharacters)
    text = traitlets.CUnicode()
    draw_instructions = traitlets.List()
    origin = traitlets.Tuple(
        traitlets.CFloat(), traitlets.CFloat(), default_value=(-1, -1)
    )
    scale = traitlets.CFloat(1.0)

    @traitlets.observe("text")
    def _observe_text(self, change):
        text = change["new"]
        lines = text.split("\n")
        draw_instructions = []
        y = 0
        for line in reversed(lines):
            x = 0
            dy = 0
            for c in line:
                e = self.data.characters[c]
                draw_instructions.append((x, y, e.texture, e.vbo_offset))
                dy = max(dy, e.vert_advance)
                x += e.hori_advance
            y += dy
        self.draw_instructions = draw_instructions

    def _set_uniforms(self, scene, shader_program):
        pass

    def draw(self, scene, program):
        viewport = np.array(GL.glGetIntegerv(GL.GL_VIEWPORT), dtype="f4")
        program._set_uniform("viewport", viewport)
        each = self.data.vertex_array.each
        for x, y, tex, vbo_offset in self.draw_instructions:
            with tex.bind(0):
                program._set_uniform("x_offset", float(x))
                program._set_uniform("y_offset", float(y))
                program._set_uniform("x_origin", self.origin[0])
                program._set_uniform("y_origin", self.origin[1])
                program._set_uniform("scale", self.scale)
                GL.glDrawArrays(GL.GL_TRIANGLES, vbo_offset * each, each)



import traitlets

from yt_idv.opengl_support import Texture, VertexArray


class SceneData(traitlets.HasTraits):
    """A class that defines a collection of GPU-managed data.

    This class contains the largest common set of features that can be used
    OpenGL rendering: a set of vertices and a set of vertex attributes.  Note
    that this is distinct from the shader, which can be swapped out and
    provided with these items.

    """

    name = traitlets.Unicode()
    vertex_array = traitlets.Instance(VertexArray)
    textures = traitlets.List(trait=traitlets.Instance(Texture))

    min_val = traitlets.CFloat(0.0)
    max_val = traitlets.CFloat(1.0)

    def _normalize_by_min_max(self, data):
        # linear normalization of data across full data range

        n_data = (data - self.min_val) / self.val_range
        n_data[n_data < 0] = 0.0
        n_data[n_data > 1] = 1.0
        return n_data

    @property
    def val_range(self):
        # the data range (max - min) across all data.

        # note: not using traitlets here because it seemed overkill to
        # observe min_val and max_val
        return self.max_val - self.min_val
